/*
 *	BOARD.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */


#include "defs.h"
#include "data.h"
#include "protos.h"
#include <stdio.h>

/* init() sets the board to the initial game state */

void init()
{
	int i;

	for (i = 0; i < 64; ++i) {
		color[i] = init_color[i];
		piece[i] = init_piece[i];
	}
	side = LIGHT;
	xside = DARK;
	last_capture = 0;
	ply = 0;
	gen_begin[0] = 0;
	undos = 0;
}


/* in_check() returns TRUE if side s is in check and FALSE
   otherwise. It just scans the board to find side s's king
   and calls attack() to see if it's being attacked. */

BOOL in_check(int s)
{
	int i;

	for (i = 0; i < 64; ++i)
		if (color[i] == s && piece[i] == KING)
			return attack(i, s ^ 1);
	return TRUE;  /* shouldn't get here */
}


/* attack() returns TRUE if square sq is being attacked by side
   s and FALSE otherwise. */

BOOL attack(int sq, int s)
{
	int i, j, n;

	for (i = 0; i < 64; ++i)
		if (color[i] == s) {
		  for (j = 0; j < offsets[piece[i]]; ++j)
		    for (n = i;;) {
		      n = mailbox[mailbox64[n] + offset[s][piece[i]][j]];
		      if (n == -1)
			break;
		      if (n == sq)
			return TRUE;
		      if (color[n] != EMPTY)
			break;
		      if (!slide[piece[i]])
			break;
		    }
		}
	return FALSE;
}


/* gen() generates pseudo-legal moves for the current position.
   It scans the board to find friendly pieces and then determines
   what squares they attack. When it finds a piece/square
   combination, it calls gen_push to put the move on the "move
   stack." */

void gen()
{
	int i, j, n;

	/* so far, we have no moves for the current ply */
	gen_end[ply] = gen_begin[ply];

	for (i = 0; i < 64; ++i)
		if (color[i] == side) {
		  for (j = 0; j < offsets[piece[i]]; ++j)
		    for (n = i;;) {
		      n = mailbox[mailbox64[n] + offset[side][piece[i]][j]];
		      if (n == -1)
			break;
		      if (color[n] != EMPTY) {
			if (color[n] == xside) {
			  if ( can_promote(i,n) ) {
			    gen_push(i, n, 17);
			    if ( ! must_promote(i,n) ) gen_push(i, n, 1);
			  } else 
			    gen_push(i,n,1);
			}
			break;
		      } else {
			if ( can_promote(i,n) ) {
			  gen_push(i, n, 16);
			  if ( ! must_promote(i,n) ) gen_push(i, n, 0);
			} else gen_push(i, n, 0);
		      }
		      if (!slide[piece[i]])
			break;
		    }
		}

	/* the next ply's moves need to start where the current
	   ply's end */
	gen_begin[ply + 1] = gen_end[ply];
}


/* gen_caps() is basically a copy of gen() that's modified to
   only generate capture and promote moves. It's used by the
   quiescence search. */

void gen_caps()
{
	int i, j, n;

	gen_end[ply] = gen_begin[ply];
	for (i = 0; i < 64; ++i)
		if (color[i] == side) {		    
		  for (j = 0; j < offsets[piece[i]]; ++j)
		    for (n = i;;) {
		      n = mailbox[mailbox64[n] + offset[side][piece[i]][j]];
		      if (n == -1) /* out of board */
			break;
		      if (color[n] != EMPTY) {
			if (color[n] == xside) /* enemy */
			  if ( can_promote(i,n) ) {
			    gen_push(i, n, 17);
			    if ( ! must_promote(i,n) ) gen_push(i, n, 1);
			  } else
 			      gen_push(i, n, 1);
			break; /**?**/
		      } else { /* empty square, add if promotion */
			if ( can_promote(i,n) ) {
			    gen_push(i, n, 16);
			}
		      }
		      if (!slide[piece[i]])
			break;
		    }
		}
	gen_begin[ply + 1] = gen_end[ply];
}

BOOL can_promote( int from, int to ) {
  int idx = can_promote_zone[side][piece[from]];
  /* printf("piece: %d, from: %d, to: %d, idx: %d\n", piece[from],from,to,idx); */

  if ( !idx ) return FALSE;
  if ( side == LIGHT ) {
    if ( ( from < idx)  || (to < idx) ) {
      /* printf("can LIGHT piece: %d, from: %d, to: %d, idx: %d\n", piece[from],from,to,idx); */
      return TRUE;
    }
  }  else { 
    if ( (from > idx) || (to > idx) ) {
      /* printf("can DARK piece: %d, from: %d, to: %d, idx: %d\n", piece[from],from,to,idx); */
      return TRUE;
    }
  }
  return FALSE;
}

BOOL must_promote( int from, int to ) {
  int idx = must_promote_zone[side][piece[from]];
  if ( !idx ) return FALSE;
  if ( side == LIGHT ) {
    if ( (from < idx) || (to < idx) ) {
      /* printf("must LIGHT piece: %d, from: %d, to: %d, idx: %d\n", piece[from],from,to,idx); */
      return TRUE;
    }
  }  else { 
    if ( (from > idx) || (to > idx) ) {
      /* printf("must DARK piece: %d, from: %d, to: %d, idx: %d\n", piece[from],from,to,idx); */
      return TRUE;
    }
  }
  return FALSE;
}



/* gen_push() puts a move on the move stack, unless it's a
   pawn promotion that needs to be handled by gen_promote().
   It also assigns a score to the move for alpha-beta move
   ordering. If the move is a capture, it uses MVV/LVA
   (Most Valuable Victim/Least Valuable Attacker). Otherwise,
   it uses the move's history heuristic value. Note that
   1,000,000 is added to a capture move's score, so it
   always gets ordered above a "normal" move. */

void gen_push(int from, int to, int bits)
{
	gen_t *g;
	
	if (bits & 16) {
	  gen_promote(from,to,bits);
	  return;
	}
	g = &gen_dat[gen_end[ply]];
	++gen_end[ply];
	g->m.b.from = (char)from;
	g->m.b.to = (char)to;
	g->m.b.promote = 0;
	g->m.b.bits = (char)bits;
	if (color[to] != EMPTY)
		g->score = 1000000 + (piece[to] * 10) - piece[from];
	else
		g->score = history[from][to];
}


/* gen_promote() is just like gen_push(), only it puts 4 moves
   on the move stack, one for each possible promotion piece */

void gen_promote(int from, int to, int bits)
{
	int i;
	gen_t *g;
	
	i = promotion[ piece[from] ];
	g = &gen_dat[gen_end[ply]];
	++gen_end[ply];
	g->m.b.from = (char)from;
	g->m.b.to = (char)to;
	g->m.b.promote = (char)i;
	g->m.b.bits = (char)(bits | 32);
	g->score = 1000000 + (i * 10);
}


/* makemove() makes a move. If the move is illegal, it
   undoes whatever it did and returns FALSE. Otherwise, it
   returns TRUE. */

BOOL makemove(move_bytes m)
{
	
	/* back up information so we can take the move back later. */
	hist_dat[ply].m.b = m;
	hist_dat[ply].oldpiece = piece[(int)m.from];
	hist_dat[ply].capture = piece[(int)m.to];
	if ( piece[(int)m.to] != EMPTY ) {
	  pieces_count[xside] --;
	  last_capture = 0;
	} else last_capture++;
	++ply;

	/* move the piece */
	color[(int)m.to] = side;
	if (m.bits & 32)
		piece[(int)m.to] = m.promote;
	else
		piece[(int)m.to] = piece[(int)m.from];
	color[(int)m.from] = EMPTY;
	piece[(int)m.from] = EMPTY;


	/* switch sides and test for legality (if we can capture
	   the other guy's king, it's an illegal position and
	   we need to take the move back) */
	side ^= 1;
	xside ^= 1;
	if (in_check(xside)) {
  	  takeback();
	  return FALSE;

	}
        /* WE HAVE SWITCHED SIDES NOW */
	/* check for bare king */

	if ( pieces_count[xside] < 2 ) { /* with only the bare king left */
	  if (( hist_dat[ply-1].capture == EMPTY ) || 
	      ( pieces_count[side] > 2 )) { /* the opponent can't be bared */
	    takeback();
	    return FALSE;
	  } 
	  if ( pieces_count[side] == 2 ) { /* opponent has only 2 pieces left */
	    if ( hist_dat[ply-1].capture != EMPTY )
	      takeback();
	  } else {
	    takeback();
	    return FALSE;
	  }
	}
	/* repeating a position is not allowed */

	if ( reps() ) {
	  takeback();
	  return FALSE;
	  }
	return TRUE;
}


/* reps() returns the number of times that the current
   position has been repeated. Thanks to John Stanback
   for this clever algorithm. */

int reps()
{
	int i;
	int b[64];
	int c = 0;  /* count of squares that are different from
				   the current position */
	int r = 0;  /* number of repetitions */

	/* is a repetition impossible? */
	if (last_capture <= 3)
		return 0;

	for (i = 0; i < 64; ++i)
		b[i] = 0;

	/* loop through the reversible moves */
	/* this move first */
	if ( ++b[hist_dat[0].m.b.from] == 0 ) --c;
	else ++c;
	if ( --b[hist_dat[0].m.b.to] == 0 ) --c;
	else ++c;
	for (i = 0; i < last_capture; i++) {
	  if ( ( undo_dat[i].m.b.from > 63 ) ||
	       ( undo_dat[i].m.b.from < 0 ) ||
	       ( undo_dat[i].m.b.to > 63 ) ||
	       ( undo_dat[i].m.b.from < 0 ) ) {
	    break;
	  }
		if (++b[undo_dat[i].m.b.from] == 0)
			--c;
		else
			++c;
		if (--b[undo_dat[i].m.b.to] == 0)
			--c;
		else
			++c;
		if (c == 0)
			++r;
	}

	return r;
}


/* takeback() is very similar to makemove(), only backwards :)  */

void takeback()
{
	move_bytes m;

	side ^= 1;
	xside ^= 1;
	--ply;
	m = hist_dat[ply].m.b;
	color[(int)m.from] = side;
	if (m.bits & 32)
		piece[(int)m.from] = hist_dat[ply].oldpiece;
	else
		piece[(int)m.from] = piece[(int)m.to];
	if (hist_dat[ply].capture == EMPTY) {
		color[(int)m.to] = EMPTY;
		piece[(int)m.to] = EMPTY;
	}
	else {
		color[(int)m.to] = xside;
		pieces_count[xside]++;
		piece[(int)m.to] = hist_dat[ply].capture;
	}
}
